// This file is required by the index.html file and will
// be executed in the renderer process for that window.

var workflowIntegration = require('./WorkflowIntegration.node');
var shell = require('electron').shell;

const path = require('path');
const https = require('https');
const fs = require('fs');

const mediaIdMetadataProp = 'Track 4';
const wolkieMarkerName = 'wolkie_marker';
const defaultMarkupColor = 'Blue';
const defaultMarkupTitle = 'Wolkie Markup';
const wolkieLocalStorageFolderListKey = 'bd_wolkie_folder_list_keys'
const frameRate = 30

var downloadRequestsArr = [];

function callAngularFunction(name, param) {
    var wolkieApp = document.querySelector('wolkie-app');
    wolkieApp.functionListener = {name, param};
}

function callCSIEvalScript(script) {
    // return new Promise((resolve) => {
    //     csInterface.evalScript(script, function(result) {
    //         resolve(isJsonString(result) ? JSON.parse(result) : result);
    //     });
    // });
}

function openUrl(url) {
    shell.openExternal(url);
}

function isJsonString(str) {
    try {
        JSON.parse(str);
    } catch (e) {
        return false;
    }
    return true;
}

function createFolderIfDoesntExists(folder) {
    if (!fs.existsSync(folder)){
        fs.mkdirSync(folder);
    }
}

function getResolveObject() {
    return workflowIntegration.GetResolve();
}

function getProject() {
    const resolveObj = getResolveObject();

    return resolveObj.GetProjectManager().GetCurrentProject();
}

function getMediaPool() {
    const resolveObj = getResolveObject();

    return resolveObj.GetProjectManager().GetCurrentProject().GetMediaPool();
}

function getMediaStorage() {
    const resolveObj = getResolveObject();
    
    return resolveObj.GetMediaStorage();
}

function getMediaMetadata(media) {
    if (!media) {
        return '';
    }

    return media.GetMetadata(mediaIdMetadataProp);
}

function getAllMediaPoolItems() {
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();
    const folders = rootFolder.GetSubFolderList();
    let mediaPoolItems = rootFolder.GetClipList();

    const getMedias = (folderList) => {
        for(let folder of folderList) {
            const newFolderList = folder.GetSubFolderList();
            mediaPoolItems = mediaPoolItems.concat(...folder.GetClipList());
            if (newFolderList.length > 0) {
                return getMedias(newFolderList);
            }
        }
        return;
    }

    getMedias(folders);


    return mediaPoolItems;
}

function getMediaPoolItemByMediaId(mediaId) {
    if (!mediaId) { return null }
    const mediaPoolList = getAllMediaPoolItems();
    let itemToReturn = mediaPoolList.find(mediaPoolItem => mediaPoolItem.GetMetadata(mediaIdMetadataProp) === mediaId);

    return itemToReturn;
}

function getAllWolkieMedias() {
    const allMedias = getAllMedias();

    return allMedias.filter((media) => (getMediaMetadata(media) !== '' && getMediaMetadata(media) !== undefined));
}

function getAllMedias() {
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();
    const folders = rootFolder.GetSubFolderList();
    const items = rootFolder.GetClipList();
    
    const getMediaItems = (subFolders = [], mediaItems = []) => {
        for(let f of subFolders) {
            const medias = f.GetClipList();
            const folders = f.GetSubFolderList();
            mediaItems = mediaItems.concat(medias);

            if (folders.length > 0) {
                mediaItems = getMediaItems(folders, mediaItems);
            }
        }

        return mediaItems;
    }

    return getMediaItems(folders, items);
}

function getMediaPathByMediaId(mediaId) {
    const media = getMediaPoolItemByMediaId(mediaId);


    return media?.GetClipProperty('File Path') ?? '';
}

function setMediaName(media, fileName) {
    media.SetClipProperty('Clip Name', fileName);

    return media;
}

function setNameByMediaId(mediaId, fileName) {
    let media = getMediaPoolItemByMediaId(mediaId);

    media = setMediaName(media, fileName);

    return JSON.stringify(media);
}

function setMetadata(media, mediaId) {
    const metadata = {};
    metadata[mediaIdMetadataProp] = mediaId;
    media.SetMetadata(metadata);
}

function setRootFolder() {
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();
    mediaPool.SetCurrentFolder(rootFolder);
}

function addNewFileToProject(filePath, mediaId, fileName, parentId) {
    const mediaStorage = getMediaStorage();
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();


    if (parentId) {
        const wolkieParentFolder = findFolderAtBdFolderListByMediaId(parentId);
        const parentFolder = getFolderByFolderId(wolkieParentFolder.folder?.folderId ?? '', rootFolder.GetSubFolderList());

        mediaPool.SetCurrentFolder(parentFolder ?? rootFolder);
    }

    let medias = mediaStorage.AddItemListToMediaPool(filePath);
    if (medias.length === 0) {
        medias = mediaPool.ImportMedia([filePath]);
    }
    
    setRootFolder();

    if (medias.length > 0) {
        setMediaName(medias[0], fileName);
        setMetadata(medias[0], mediaId)
        
        return JSON.stringify(medias[0]);
    }

    return JSON.stringify({});
}

function attachProxy(mediaId, filePath, fileName, highRes = 0) {
    let media = getMediaPoolItemByMediaId(mediaId);
    let pluginProxyError = false;

    if (media) {
        if (media.GetClipProperty('Proxy').toLowerCase() === 'none') {
            let proxyPath = filePath;
            if (highRes) {
                proxyPath = media.GetClipProperty('File Path');
                const replaced = media.ReplaceClip(filePath);

            }

            const proxy = media.LinkProxyMedia(proxyPath);

            if (!proxy) {
                pluginProxyError = true;
            }
        }
        media = setMediaName(media, fileName);
        setMetadata(media, mediaId);

        setRootFolder();

        return JSON.stringify({...media, pluginProxyError});
    }

    return null;
}

function clearMarkers(mediaId) {
    const media = getMediaPoolItemByMediaId(mediaId);
    if (media) {
        const markersObj = media.GetMarkers();
        const keys = Object.keys(markersObj);
        const markers = keys.map((key) => markersObj[key]).filter((marker) => marker.customData === wolkieMarkerName);
        for(let i = 0; i < markers.length; i++) {
            media.DeleteMarkerByCustomData(wolkieMarkerName);
        }
    }
}

function getFrame(seconds) {
    return Math.floor(seconds * frameRate);
}

function createMarker(mediaId, comments, time) {
    const media = getMediaPoolItemByMediaId(mediaId);

    if (media) {
        const frameId = getFrame(time)
        const resp = media.AddMarker(frameId, defaultMarkupColor, defaultMarkupTitle, `${comments}`, 1, wolkieMarkerName);
    }
}

function createFolder(folderName, parentFolder) {
    const mediaPool = getMediaPool();

    if (!parentFolder) {
        parentFolder = mediaPool.GetRootFolder();
    }
    
    return mediaPool.AddSubFolder(parentFolder, folderName);
}


function getFolderByFolderId(folderId, folderList, idx = 0) {
    folder = null;
    for (let f of folderList) {
        if (f.uniqueId === folderId) {
            folder = f;
            break;
        } else {
            folder = getFolderByFolderId(folderId, f.GetSubFolderList());
            if (folder) {
                break;
            }
        }
    }

    return folder;
}

function moveFolder(folders, folderTarget) {
    const mediaPool = getMediaPool();
    if (!folderTarget) {
        folderTarget = mediaPool.GetRootFolder();
    }

    mediaPool.MoveFolders(folders, folderTarget);
}

function moveMedia(mediaId, parentIdTarget) {
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();
    const media = getMediaPoolItemByMediaId(mediaId);
    const wolkieParentFolder = findFolderAtBdFolderListByMediaId(parentIdTarget);
    const targetFolder = getFolderByFolderId(wolkieParentFolder.folder?.folderId, rootFolder.GetSubFolderList());

    mediaPool.MoveClips([media], targetFolder ?? rootFolder);
}

function clearDatabase() {
    window.localStorage.setItem(wolkieLocalStorageFolderListKey, '[]');
}

function renameFolder(newName, folder, folderParent = null) {
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();

    if (!folderParent) {
        folderParent = rootFolder;
    }

    const newFolder = createFolder(newName, folderParent);
    const oldSubFolders = folder.GetSubFolderList();
    const oldClips = folder.GetClipList();

    mediaPool.SetCurrentFolder(folder);

    moveFolder(oldSubFolders, newFolder);
    if (oldClips.length > 0) {
        mediaPool.MoveClips(oldClips, newFolder);
    }
    mediaPool.SetCurrentFolder(rootFolder);
    mediaPool.DeleteFolders([folder]);

    return newFolder;
}

function createOrUpdateFolder(mediaId, name, parentId) {
    const mediaPool = getMediaPool();
    const rootFolder = mediaPool.GetRootFolder();
    const rootFolderList = rootFolder.GetSubFolderList();
    const wolkieParentFolder = findFolderAtBdFolderListByMediaId(parentId);
    const wolkieMediaFolder = findFolderAtBdFolderListByMediaId(mediaId);    
    
    let mediaFolder = null;
    let parentFolder = rootFolder;

    if (wolkieParentFolder.folder) {
        parentFolder = getFolderByFolderId(wolkieParentFolder.folder.folderId, rootFolderList);

        if (!parentFolder) {
            const newParentFolder = getFolderByFolderId(wolkieMediaFolder.folder.parentFolderId, rootFolderList);
            parentFolder = createFolder(wolkieParentFolder.folder.name, newParentFolder ?? rootFolder);
        }
    }    

    if (wolkieMediaFolder.folder) {
        mediaFolder = getFolderByFolderId(wolkieMediaFolder.folder.folderId, rootFolderList);
        if (!mediaFolder) {
            mediaFolder = createFolder(name, parentFolder);
        } else {
            if (mediaFolder.GetName() !== name) {
                mediaFolder = renameFolder(name, mediaFolder, parentFolder);
            }
            moveFolder([mediaFolder], parentFolder);
        }
    } else {
        mediaFolder = createFolder(name, parentFolder);
    }

    addOrUpdateFolderAtBdFolderList(name, mediaFolder.uniqueId, mediaId, parentFolder.uniqueId);
}

function getFolderList() {
    const folderListStringfied = window.localStorage.getItem(wolkieLocalStorageFolderListKey);

    return JSON.parse(folderListStringfied ?? '[]');
}

function setFolderList(list) {
    window.localStorage.setItem(wolkieLocalStorageFolderListKey, JSON.stringify(list));
}

function addOrUpdateFolderAtBdFolderList(name, folderId, mediaId, parentFolderId) {
    const folderList = getFolderList();
    const {folder, idx} = findFolderAtBdFolderListByMediaId(mediaId);

    if (folder && idx >= 0) {
        folder.name = name;
        folder.parentFolderId = parentFolderId;
        folder.folderId = folderId;
        folderList[idx] = folder
    } else {
        const newFolder = {name, folderId, mediaId, parentFolderId};
        folderList.push(newFolder);
    }

    setFolderList(folderList);
}

function findFolderAtBdFolderListByMediaId(mediaId) {
    const folderList = getFolderList();
    let folder = null;
    let idx = -1;

    if (folderList.length > 0) {
        for(let i = 0; i < folderList.length; i++) {
            if (folderList[i].mediaId === mediaId) {
                folder = folderList[i];
                idx = i;
                break;
            }
        }
    }

    return {folder, idx};
}

async function downloadFileToProject(reviewModel, downloadType, updateMarkups, progressFn, completeFn) {
    const resolveObj = getResolveObject();
    const proxy = `${reviewModel.media.workspaceId}/${reviewModel.media.id}`;
    
    setRootFolder();
    
    let projectPath = resolveObj.GetMediaStorage().GetMountedVolumeList()[0];

    let downloadDirectory = `${projectPath}${path.sep}wolkie-downloads`;
    createFolderIfDoesntExists(downloadDirectory);
    const folders = proxy.split('/');
    downloadDirectory += `${path.sep}${folders[0]}`;
    createFolderIfDoesntExists(downloadDirectory);
    downloadDirectory += `${path.sep}${folders[1]}`;
    createFolderIfDoesntExists(downloadDirectory);

    var originalNameArr = reviewModel.media.name.split(`.`);
    var originalExtension = originalNameArr[originalNameArr.length -1];
    var fileName = reviewModel.media.name.replace(/[/\\?%*:|"<>]/g, '-');
    fileName = fileName.split(`.`);
    if (fileName.length > 1) {
        fileName.pop();
    }
    fileName = fileName.join(``);

    var originalInsertionName = `wolkie_${proxy.replace('/', '_')}.${originalExtension}`;

    var originalFullPath = `${downloadDirectory}${path.sep}${originalInsertionName}`.replace(/\\/g, "\\");
    var originalFileName = `${fileName}.${originalExtension}`;
    var originalUrl = reviewModel.files.find(f => f.quality === `original`).url;

    var proxyFullPath = '';
    var proxyInsertionName = '';
    var proxyFileName = '';
    var proxyUrl = '';

    if (reviewModel.media.mediaType === 'video') {
        proxyFileName = `${fileName}.mp4`;
        proxyInsertionName = `wolkie_${proxy.replace('/', '_')}_proxy.mp4`;
        proxyUrl = reviewModel.files.find(f => f.quality === `360p`).url || ``;
        proxyFullPath = `${downloadDirectory}${path.sep}${proxyInsertionName}`.replace(/\\/g, "\\");
    } else {
        downloadType = 'ORIGINAL_ONLY';
    }

    var canceled = false;
    var cancelFn = () => {
        canceled = true;
    };
    var total = 0;
    var progress = 0;

    var currentFilePath = getMediaPathByMediaId(reviewModel.media.id);
    const {folder} = findFolderAtBdFolderListByMediaId(reviewModel.media.id);
    const needToMoveMedia = folder?.folderId !== reviewModel.media.parentId;
    switch (downloadType) {
        case 'ORIGINAL_ONLY':
            await downloadFileToPath(
                reviewModel.media.id,
                originalFullPath, 
                originalUrl, 
                (chunkSize) => {
                    progress += chunkSize;
                    progressFn(progress, total);
                },
                (totalSize) => {
                    total += totalSize;
                    progressFn(progress, total);
                },
                cancelFn
            );
            if (!canceled) {
                if (currentFilePath) {
                    if (needToMoveMedia) {
                        moveMedia(reviewModel.media.id, reviewModel.media.parentId);
                    }
                    if (currentFilePath.includes('proxy')) {
                        var projectItem = attachProxy(reviewModel.media.id, originalFullPath, originalFileName, 1);
                        completeFn(projectItem);
                    } else {
                        var projectItem = setNameByMediaId(reviewModel.media.id, originalFileName);
                        completeFn(projectItem);
                    }
                } else {
                    var projectItem = addNewFileToProject(originalFullPath, reviewModel.media.id, originalFileName, reviewModel.media.parentId);
                    completeFn(projectItem);
                }
            }
            break;
        case 'PROXY_ONLY':
            await downloadFileToPath(
                reviewModel.media.id,
                proxyFullPath, 
                proxyUrl, 
                (chunkSize) => {
                    progress += chunkSize;
                    progressFn(progress, total);
                },
                (totalSize) => {
                    total += totalSize;
                    progressFn(progress, total);
                },
                cancelFn
            );

            if (!canceled) {
                if (currentFilePath) {
                    if (moveMedia) {
                        moveMedia(reviewModel.media.id, reviewModel.media.parentId);
                    }
                    if (currentFilePath.includes('proxy')) {
                        var projectItem = setNameByMediaId(reviewModel.media.id, proxyFileName);
                        completeFn(projectItem);
                    } else {
                        var projectItem = attachProxy(reviewModel.media.id, proxyFullPath, originalFileName);
                        completeFn(projectItem);
                    }
                } else {
                    var projectItem = addNewFileToProject(proxyFullPath, reviewModel.media.id, proxyFileName, reviewModel.media.parentId);
                    completeFn(projectItem);
                }
            }
            break;
        case 'ALL':
            var downloadPromises = [
                downloadFileToPath(
                    reviewModel.media.id,
                    originalFullPath, 
                    originalUrl, 
                    (chunkSize) => {
                        progress += chunkSize;
                        progressFn(progress, total);
                    },
                    (totalSize) => {
                        total += totalSize;
                        progressFn(progress, total);
                    },
                    cancelFn
                )
            ];

            if (reviewModel.media.mediaType === 'video') {
                downloadPromises.push(downloadFileToPath(
                    reviewModel.media.id,
                    proxyFullPath, 
                    proxyUrl, 
                    (chunkSize) => {
                        progress += chunkSize;
                        progressFn(progress, total);
                    },
                    (totalSize) => {
                        total += totalSize;
                        progressFn(progress, total);
                    },
                    cancelFn
                ));
            }

            await Promise.all(downloadPromises);

            if (!canceled) {
                if (currentFilePath) {
                    if (moveMedia) {
                        moveMedia(reviewModel.media.id, reviewModel.media.parentId);
                    }
                    if (currentFilePath.includes('proxy')) {
                        var projectItem = attachProxy(reviewModel.media.id, originalFullPath, originalFileName);
                        completeFn(projectItem);
                    } else {
                        var projectItem = attachProxy(reviewModel.media.id, proxyFullPath, originalFileName);
                        completeFn(projectItem);
                    }
                } else {
                    var projectItem = addNewFileToProject(originalFullPath, reviewModel.media.id, originalFileName, reviewModel.media.parentId);
                    if (reviewModel.media.mediaType === 'video') {
                        attachProxy(reviewModel.media.id, proxyFullPath, originalFileName);
                    }
                    completeFn(projectItem);
                }
            }
            break;
        default:
            break;
    }

    downloadRequestsArr = downloadRequestsArr.filter(d => d.id !== reviewModel.media.id);

    if (!canceled) {
        if (updateMarkups) {
            clearMarkers(reviewModel.media.id);
            var markerModels = [];

            reviewModel.markups.forEach(markup => {
                var time = markup.specs.time;
                var color = markup.color
                var comments = markup.comments.map(c => {
                    return `${c.createdByAvatar.name}(${new Date(c.createdAt).toLocaleString()}): ${c.text}`;
                }).join(`\\n`);
                var idx = markerModels.findIndex(mm => mm?.specs?.time === markup.specs.time);
                if (idx === -1) {
                    markerModels.push({time, comments});
                } else {
                    markerModels[idx].comments += `\n\n${comments}`;
                }
            });
            markerModels.forEach(async m => {
                createMarker(reviewModel.media.id, m.comments, m.time);
            });
        }
    }
}

async function downloadFileToPath(mediaId, filePath, url, progressFn, totalFn, onCancelFn) {
    return new Promise((resolve) => {
        if (fs.existsSync(filePath)) {
            var fileSize = getFilesizeInBytes(filePath);
            if (totalFn) {
                totalFn(fileSize);
            }
            if (progressFn) {
                progressFn(fileSize);
            }
            resolve(null);
        } else {    
            var file = fs.createWriteStream(filePath);
            var totalBytes = 0;
            var request = https.get(url)
                        .on(`response`, (res) => {
                            totalBytes = +res.headers['content-length'];
                            if (totalFn) {
                                totalFn(totalBytes);
                            }
                            
                            file.on("finish", () => {
                                file.close();
                                file = null;
                            });
                            res
                                .on('data', (chunk) => {
                                    if (progressFn) {
                                        progressFn(chunk.length);
                                    }
                                })
                                .on('end', () => {
                                    request = null;
                                    resolve(null);
                                })
                                .pipe(file);
                        });
            var cancelFn = () => {
                if (onCancelFn) {
                    onCancelFn();
                }
                resolve(null);
            }
            downloadRequestsArr.push({mediaId, file, request, filePath, cancelFn});
        }

    });
}

function clearDirectory(dir) {
    if (fs.existsSync(dir)) {
        fs.readdirSync(dir).forEach(f => fs.unlinkSync(`${dir}${path.sep}${f}`));
    }
}

function cancelDownload(mediaId) {
    var items = downloadRequestsArr.filter(d => d.id === mediaId);
    
    items.forEach(item => {
        if (item.request) {
            item.request.abort();
        }
        
        if (item.file && item.file.fd) {
            fs.closeSync(item.file.fd);
        }

        if (fs.existsSync(item.filePath)) {
            fs.unlinkSync(item.filePath);
        }

        if (item.cancelFn) {
            item.cancelFn();
        }
    });
}

function getFilesizeInBytes(path) {
    var stats = fs.statSync(path);
    return stats.size;
}

function jsxError(response) {
    return typeof response === `string` && (response === `EvalScript error.` || response.toLowerCase().includes(`evalscript error`));
}

async function createOrUpdateFolderTree(medias) {
    const folderTree = async (media) => {
        createOrUpdateFolder(media.id, media.name, media.parentId);
        for await (var m of media.children) {
            await folderTree(m);
        }
    };
    
    for await (var m of medias) {
        await folderTree(m);
    }
}

async function getAllFilesMediaIdsFromDavinciLibrary() {
    const project = getProject();
    const medias = getAllWolkieMedias();
    const mediaIds = medias.map((media) => getMediaMetadata(media));

    if (mediaIds && typeof mediaIds === `string`) {
        mediaIds = mediaIds.split(`,`);
    }
    return mediaIds || [];
}

